#ifndef H_BLE_KEYBOARD_
#define H_BLE_KEYBOARD_

#include "nimble/ble.h"
#include "modlog/modlog.h"

#include "host/ble_gatt.h"
#include "host/ble_uuid.h"
#include "host/ble_hs_pvcy.h"
#include "host/util/util.h"
#include "services/gap/ble_svc_gap.h"
#include "services/gatt/ble_svc_gatt.h"
#include "services/bas/ble_svc_bas.h"
//#include "services/dis/ble_svc_dis.h"
#include "os/os_mbuf.h"

#include "nimble/nimble_port.h"
#include "nimble/nimble_port_freertos.h"
#include "host/ble_hs.h"

//////////////////////////////////////////////////
// デバイス情報サービス　Device Infomation Service
//////////////////////////////////////////////////

// デバイス情報サービス用UUID
#define GATT_UUID_DIS_SERVICE             0x180A
#define GATT_UUID_DIS_SYSTEM_ID           0x2A23
#define GATT_UUID_DIS_MODEL_NUMBER        0x2A24
#define GATT_UUID_DIS_SERIAL_NUMBER       0x2A25
#define GATT_UUID_DIS_FIRMWARE_REVISION   0x2A26
#define GATT_UUID_DIS_HARDWARE_REVISION   0x2A27
#define GATT_UUID_DIS_SOFTWARE_REVISION   0x2A28
#define GATT_UUID_DIS_MANUFACTURER_NAME   0x2A29
#define GATT_UUID_DIS_PNP_INFO            0x2A50


//////////////////////////////////////////////////
// characteristic presentation format
//////////////////////////////////////////////////
// 名前空間定数（現状は下記の1種類だけ）
#define GATT_NAME_SPACE_BLUETOOTH_SIG     0x01

// キャラクタリスティックで読み取ったときの値のフォーマット
#define GATT_CPF_FORMAT_RFU               0x00
#define GATT_CPF_FORMAT_BOOLEAN           0x01
#define GATT_CPF_FORMAT_2BIT              0x02
#define GATT_CPF_FORMAT_NIBBLE            0x03
#define GATT_CPF_FORMAT_UINT8             0x04
#define GATT_CPF_FORMAT_UINT12            0x05
#define GATT_CPF_FORMAT_UINT16            0x06
#define GATT_CPF_FORMAT_UINT24            0x07
#define GATT_CPF_FORMAT_UINT32            0x08
#define GATT_CPF_FORMAT_UINT48            0x09
#define GATT_CPF_FORMAT_UINT64            0x0A
#define GATT_CPF_FORMAT_UINT128           0x0B
#define GATT_CPF_FORMAT_SINT8             0x0C
#define GATT_CPF_FORMAT_SINT12            0x0D
#define GATT_CPF_FORMAT_SINT16            0x0E
#define GATT_CPF_FORMAT_SINT24            0x0F
#define GATT_CPF_FORMAT_SINT32            0x10
#define GATT_CPF_FORMAT_SINT48            0x11
#define GATT_CPF_FORMAT_SINT64            0x12
#define GATT_CPF_FORMAT_SINT128           0x13
#define GATT_CPF_FORMAT_FLOAT32           0x14
#define GATT_CPF_FORMAT_FLOAT64           0x15
#define GATT_CPF_FORMAT_SFLOAT            0x16
#define GATT_CPF_FORMAT_FLOAT             0x17
#define GATT_CPF_FORMAT_DUINT16           0x18
#define GATT_CPF_FORMAT_UTF8S             0x19
#define GATT_CPF_FORMAT_UTF16S            0x1A
#define GATT_CPF_FORMAT_STRUCT            0x1B

// キャラクタリスティックの値を説明するための情報
typedef struct characteristic_presentation_format{
  uint16_t  unit;         // UUIDで単位を表します。 3.5 Units を参照
  uint16_t  description;  // name_space=1の時は　2.4.2.1 Bluetooth SIG GATT Characteristic Presentation Format Description　を参照
  uint8_t   format;       // GATT_CPF_FORMAT_〜定数で選択。数値のフォーマットを表す
  uint8_t   exponent;     // 指数。キャラクタリスティックの値に10の何乗をかけるのか。
  uint8_t   name_space;   // descriptionの名前空間。 0：なし。 1：Bluetooth SIGによる定義　のどちらか。通常は1 
} gatt_cpf_struct;

// https://www.bluetooth.com/specifications/assigned-numbers/
// 2.4.2.1 Bluetooth SIG GATT Characteristic Presentation Format Description 抜粋
// 0x0000：不明
// 0x0001〜0x00FF：1〜255番目　という意味
// 0x0100：フロント
// 0x010F：内側
// のようになっていて、例えば温度計を設置したあとに、１番目や、内側などの意味を持たせるための説明に使用する
// 3.5 Units 抜粋
// 0x272F：温度℃
// 0x2704：電流A
// 0x2728：電圧V
// 0x27B0：電池容量Ah
// 0x2727：電荷量C
// 0x27AD：パーセント%
// 0x27AB：電力量kWh

//////////////////////////////////////////////////
// HID
//////////////////////////////////////////////////

// HIDの最新バージョン 1.11
#define HID_USB_VERSION                   0x0111

// HID protocol mode values
#define HID_PROTOCOL_MODE_BOOT            0x00      // ブートプロトコルモード
#define HID_PROTOCOL_MODE_REPORT          0x01      // レポートプロトコルモード(デフォルト。通常時はこちら)

// HIDコントロールポイントの値　下位4ビット　上位4ビットは1 = HID_CONTROL request
// Bluetooth SIGのHuman Interface Device Profile 1.0のSection 7.4.2を参照
#define HID_CTRL_POINT_NOP                0   // No Operation
#define HID_CTRL_POINT_HARD_RESET         1   // ハードウェアリセット
#define HID_CTRL_POINT_SOFT_RESET         2   // ソフトウェアリセット
#define HID_CTRL_POINT_SUSPEND            3   // 休止状態へ移行
#define HID_CTRL_POINT_EXIT_SUSPEND       4   // 休止状態から復帰
#define HID_CTRL_POINT_VIRTUAL_UNPLUG     5   // 仮想的にケーブルを引っこ抜く

// HID information flags
#define HID_FLAGS_REMOTE_WAKE             0x01      // RemoteWake
#define HID_FLAGS_NORMALLY_CONNECTABLE    0x02      // NormallyConnectable

// HID Report types
#define HID_REPORT_TYPE_INPUT             1
#define HID_REPORT_TYPE_OUTPUT            2
#define HID_REPORT_TYPE_FEATURE           3

// Bootモード時のレポートID
#define HID_BOOT_REPORT_ID_KEYBOARD       1
#define HID_BOOT_REPORT_ID_MOUSE          2

// Bootモード時のレポートサイズ
#define HID_BOOT_REPORT_SIZE_KEYBOARD     9 // 1-byte Report ID + standard 8-byte keyboard boot report
#define HID_BOOT_REPORT_SIZE_MOUSE        4 // 1-byte Report ID + standard 3-byte mouse boot report

// HIDサービス用UUID
#define GATT_UUID_HID_SERVICE             0x1812
#define GATT_UUID_HID_INFORMATION         0x2A4A
#define GATT_UUID_HID_REPORT_MAP          0x2A4B
#define GATT_UUID_HID_CONTROL_POINT       0x2A4C
#define GATT_UUID_HID_REPORT              0x2A4D
#define GATT_UUID_HID_PROTOCOL_MODE       0x2A4E

#define GATT_UUID_HID_BT_KB_INPUT         0x2A22
#define GATT_UUID_HID_BT_KB_OUTPUT        0x2A32
#define GATT_UUID_HID_BT_MOUSE_INPUT      0x2A33

// Keyboard レポートサイズ (レポートマップで自分で定義したサイズ)
#define HID_REPORT_SIZE_KEYBOARD_IN       8   // キーコードを送る時のサイズ
#define HID_REPORT_SIZE_KEYBOARD_OUT      1   // NumLockなどのLED情報を受け取る時のサイズ

// HID レポートモード用レポートID (レポートマップで自分で定義したID)
enum hid_report_id{
  HID_REPORT_ID_KB_OUT,    // LED output report ID from report map
  HID_REPORT_ID_KB_IN,     // Keyboard input report ID from report map
  HID_REPORT_ID_CC_IN,     // Consumer Control input report ID from report map
  HID_REPORT_ID_FEATURE    // Feature report ID from report map
};

// ディスクリプタ用UUID
#define GATT_UUID_BAT_PRESENT_DESCR       0x2904
#define GATT_UUID_EXT_RPT_REF_DESCR       0x2907
#define GATT_UUID_RPT_REF_DESCR           0x2908


enum attr_handles_list{
  HANDLE_BATTERY_LEVEL,

  HANDLE_DIS_MODEL_NUMBER,
  HANDLE_DIS_SERIAL_NUMBER,
  HANDLE_DIS_HARDWARE_REVISION,
  HANDLE_DIS_FIRMWARE_REVISION,
  HANDLE_DIS_SOFWARE_REVISION,
  HANDLE_DIS_MANUFACTURER_NAME,
  HANDLE_DIS_SYSTEM_ID,
  HANDLE_DIS_PNP_INFO,

  HANDLE_HID_INFORMATION,
  HANDLE_HID_CONTROL_POINT,
  HANDLE_HID_REPORT_MAP,
  HANDLE_HID_PROTOCOL_MODE,
  HANDLE_HID_KB_IN_REPORT,
  HANDLE_HID_KB_OUT_REPORT,
  HANDLE_HID_KB_CCC_DISC,
  HANDLE_HID_CC_REPORT,
  HANDLE_HID_BOOT_KB_IN_REPORT,
  HANDLE_HID_BOOT_KB_OUT_REPORT,
  HANDLE_HID_FEATURE_REPORT,

  HANDLE_NUM
};

#define MAC2STR_REV(a) (a)[5], (a)[4], (a)[3], (a)[2], (a)[1], (a)[0]

int ble_init_server();
bool ble_is_connected();
void send_text(const char *text);

void ble_store_config_init(void);

#endif
